;*********************************************************
;**                                                     **
;**        USB R/C Transmitter to PC Interface          **
;**                                                     **
;**                M. Jervis & M. Crabbe                **
;**                     April 2006                      **
;**                                                     **
;**        Original code developed by Alessio &         **
;**      Andrea, based on Prof Bradley A. Minch's       **
;** Firmware (http://pe.ece.olin.edu/ece/projects.html) **
;**                                                     **
;**   Developed and ported to the 18F2550 Microchip     **
;**                         By                          **
;**           Mark Jervis and Malcolm Crabbe            **
;**             to support 8 full analogue              **
;**                      channels                       **
;**                                                     **
;**              Development assisted by                **
;**                                                     **
;**   Thomas Black, Luke Middleton and Carl Mayrath     **
;**                                                     **
;**      Following in the spirit of the original        **
;**  developers we make these files freely available    **
;**    for distribution, provided this credit header    **
;**  remains intact in the ASM file, and that all files **
;**   are distributed in hole,  Any commercial used is  **
;**   STRICTLY forbidden, without written concent from  **
;**               the above contributors.               **
;**               contact can be made via               **
;**       http://www.rcgroups.com/forums/index.php      **
;**      Users Malc C, MarkJ, Mr RC-Cam, Allesio70      **
;**                 Lukez and Cinnamon                  **
;**                                                     **
;**    Anyone whishing to host these files on their     **
;**  website may do so, provided there is reference to  **
;**           the contributors stated above             **
;*********************************************************

#include "p18F2550.inc"
#include "usb_defs.inc"
#include "ENGR2210.inc"
#include "rcjoy_defs.inc"

; CONFIG PLLDIV=1, CPUDIV=OSC3_PLL4, USBDIV=2
; CONFIG FOSC=XTPLL_XT, FCMEM=OFF, IESO=OFF
 CONFIG PLLDIV=5, CPUDIV=OSC3_PLL4, USBDIV=2
 CONFIG FOSC=HSPLL_HS, FCMEN=OFF, IESO=OFF
 CONFIG PWRT=OFF, BOR=ON, BORV=3, VREGEN=ON
 CONFIG WDT=OFF, WDTPS=32768
 CONFIG MCLRE=ON, LPT1OSC=OFF, PBADEN=OFF, CCP2MX=ON
 CONFIG STVREN=ON, LVP=OFF, XINST=OFF, DEBUG=OFF ; ICPRT=OFF, 
 CONFIG CP0=OFF, CP1=OFF, CP2=OFF, CPB=OFF, CPD=OFF
 CONFIG WRT0=OFF, WRT1=OFF, WRT2=OFF
 CONFIG WRTB=OFF, WRTC=OFF, WRTD=OFF
 CONFIG EBTR0=OFF, EBTR1=OFF, EBTR2=OFF
 CONFIG EBTRB=OFF
 IFDEF __18F2550
 CONFIG WRT3=OFF, EBTR3=OFF
 ENDIF



;#define	SHOW_ENUM_STATUS



BANKED_DATA		udata		0x060	; on bank0 outside access memory

USB_buffer_desc		res	4
USB_buffer_data		res	8
USB_error_flags		res 1
USB_curr_config		res	1
USB_device_status	res	1
USB_protocol		res	1
USB_idle_rate		res	1
USB_dev_req			res	1
USB_address_pending	res	1
USB_desc_ptr		res	1
USB_bytes_left		res	1
USB_loop_index		res 1
USB_packet_length	res	1
USB_USTAT			res	1
USB_USWSTAT			res	1
COUNTER				res	1
LED_states			res	1



ACCESS_DATA		udata_acs

BUFFER					res 8	; location for channels data to be sent via USB
TMRBUFFER				res .24	; location for channels timer data, filled by PpmTrigger
PCLATH_save				res 1	; for ISR context saving (PCLATH)
PCLATU_save				res 1	; for ISR context saving (PCLATU)
FSR0L_save				res 1	; for ISR context saving (FSR0L)
FSR0H_save				res 1	; for ISR context saving (FSR0H)
TMR1L_save				res	1	; for saving TMR1L in PpmTrigger
TMR1H_save				res 1	; for saving TMR1H in PpmTrigger
PIR1_save				res 1	; for saving PIR1
nChPulse				res 1	; number of channels pulse received in PpmTrigger (ISR)
RefL					res 1	; store timer reference for the TmrToChData rescaling algorithm
RefH					res 1	; Ref high byte
StepL					res 1	; store step length for the TmrToChData rescaling algorithm
StepH					res 1	; Step high byte
StepNumber				res 1	; store step number for the TmrToChData rescaling loop
ChDataL					res 1	; output of TmrToChData
ChDataH					res 1	; ChData high byte
TmrL					res 1	; input of TmrToChData
TmrH					res 1	; Tmr high byte
nChannel				res 1	; channel number in ProcessChannel
nAssignment				res 1	; assignment of the current channel number in ProcessChannel
SignalPresent			res 1	; how much 10,92ms Main Loop with TMR1 overflowed before considering signal absent



STARTUP		code		0x0000

			goto		Main					; Reset vector
			nop
			nop
			goto		ISR						; High-priority interrupt vector trap
			nop
			nop
			nop
			nop
			nop
			nop
			goto		$						; Low-priority interrupt vector trap



USBDESCRIPTORS	code_pack

Descriptor
			movlw		upper Descriptor_begin
			movwf		TBLPTRU, ACCESS
			movlw		high Descriptor_begin
			movwf		TBLPTRH, ACCESS
			movlw		low Descriptor_begin
			banksel		USB_desc_ptr
			addwf		USB_desc_ptr, W, BANKED
			ifset STATUS, C, ACCESS
				incf		TBLPTRH, F, ACCESS
				ifset STATUS, Z, ACCESS
					incf		TBLPTRU, F, ACCESS
				endi
			endi
			movwf		TBLPTRL, ACCESS
			tblrd*
			movf		TABLAT, W
			return

Descriptor_begin
Device
			db			0x12				; bLength (length of this descriptor)
			db			DEVICE				; bDescriptorType
			db			0x10, 0x01			; bcdUSB (low byte, high byte; revision 1.10)
			db			0x00				; bDeviceClass (zero means each interface operates idepedently; for HID must be specified in the interface descriptor, not here)
			db			0x00				; bDeviceSubClass (for HID must be specified in the interface descriptor, not here)
			db			0x00				; bDeviceProtocl
			db			0x08				; bMaxPacketSize
			db			0xd8, 0x04			; idVendor (low byte, high byte; 0x04D8 is Microchip)
			db			0x03, 0x01			; idProduct (low byte, high byte)
			db			0x00, 0x02			; bcdDevice (low byte, high byte; 0xJJMN: JJ.M.N: JJ major version number, M minor, N sub-minor: 0x0200 = 2.0.0)
			db			0x01				; iManufacturer (manufacturer string index)
			db			0x02				; iProduct (product string index)
			db			0x00				; iSerialNumber (none)
			db			NUM_CONFIGURATIONS	; bNumConfigurations

#define		REPORT1_LENGTH		String0-Report1

Configuration1
			db			0x09					; bLength (length of this descriptor)
			db			CONFIGURATION			; bDescriptorType
			db			0x22, 0x00				; wTotalLength (low byte, high byte)
			db			NUM_INTERFACES			; bNumInterfaces
			db			0x01					; bConfigurationValue
			db			0x00					; iConfiguration (string index for this config = none)
			db			0x80					; bmAttributes (bus powered, no remote wakeup)
			db			0x32					; bMaxPower (self-powered draws 0 mA from the bus)
Interface1
			db			0x09					; bLength (length of this descriptor)
			db			INTERFACE				; bDescriptorType
			db			0x00					; bInterfaceNumber (number of interface, 0 based array)
			db			0x00					; bAlternateSetting
			db			0x01					; bNumEndpoints (number of endpoints used in this interface excluding EP0)
			db			0x03					; bInterfaceClass (always 3 for HID devices)
			db			0x00					; bInterfaceSubClass (for HID: boot device; in this case not a boot device: not a keyboard or mouse)
			db			0x00					; bInterfaceProtocol (Keyboard protocol)
			db			0x00					; iInterface (index of string descriptor describing this interface = none)
HID1
			db			0x09					; bLength (length of this descriptor)
			db			HID						; bDescriptorType
			db			0x00, 0x01				; bcdHID (low byte, high byte; HID release number = 1.00)
			db			0x00					; bCountryCode (none)
			db			0x01					; bNumDescriptors (number of HID descriptor to follow)
			db			REPORT					; bDescriptorType
			db			REPORT1_LENGTH, 0x00	; wDescriptorLength (low byte, high byte)
Endpoint1
			db			0x07					; bLength (length of this descriptor)
			db			ENDPOINT				; bDescriptorType
			db			0x81					; bEndpointAddress (EP1, IN)
			db			0x03					; bmAttributes (Interrupt)
			db			0x08, 0x00				; wMaxPacketSize (low byte, high byte; = 8)
			db			0x0A					; bInterval (polling interval = 10 ms)

Report1
			db			0x05, 0x01			;	USAGE_PAGE (Generic Desktop)
			db			0x09, 0x04			;	USAGE (Joystick)
			db			0xa1, 0x01			;	COLLECTION (Application)
			db			0x09, 0x01			;		USAGE (Pointer)
			db			0xa1, 0x00			;		COLLECTION (Physical)

			db			0x09, 0x30			;			USAGE (X)
			db			0x09, 0x31			;			USAGE (Y)
			db			0x09, 0x32			;			USAGE (Z)
			db			0x15, 0x00			;			LOGICAL_MINIMUM (0)
			db			0x26, 0xff, 0x00	;			LOGICAL_MAXIMUM (255)
			db			0x75, 0x08			;			REPORT_SIZE (8)	
			db			0x95, 0x03			;			REPORT_COUNT (3)
			db			0x81, 0x02			;			INPUT (Data,Var,Abs)

			db			0xc0				;		END_COLLECTION

			db			0x09, 0x01			;		USAGE (Pointer)
			db			0xa1, 0x00			;		COLLECTION (Physical)

			db			0x09, 0x33			;			USAGE (Rx)
			db			0x09, 0x34			;			USAGE (Ry)
			db			0x09, 0x35			;			USAGE (Rz)
			db			0x15, 0x00			;			LOGICAL_MINIMUM (0)
			db			0x26, 0xff, 0x00	;			LOGICAL_MAXIMUM (255)
			db			0x75, 0x08			;			REPORT_SIZE (8)	
			db			0x95, 0x03			;			REPORT_COUNT (3)
			db			0x81, 0x02			;			INPUT (Data,Var,Abs)

			db			0xc0				;		END_COLLECTION

			db			0x09, 0x01			;		USAGE (Pointer)
			db			0xa1, 0x00			;		COLLECTION (Physical)

			db			0x09, 0x36			;			USAGE (Slider)
			db			0x09, 0x37			;			USAGE (Dial)
			db			0x15, 0x00			;			LOGICAL_MINIMUM (0)
			db			0x26, 0xff, 0x00	;			LOGICAL_MAXIMUM (255)
			db			0x75, 0x08			;			REPORT_SIZE (8)	
			db			0x95, 0x02			;			REPORT_COUNT (2)
			db			0x81, 0x02			;			INPUT (Data,Var,Abs)

			db			0xc0				;		END_COLLECTION

			db			0xc0				;	END_COLLECTION

String0
			db			String1-String0, STRING	; bLength, bDescriptorType
			db			0x09, 0x04			; wLANGID[0] (low byte), wLANGID[0] (high byte)
String1
			db			String2-String1, STRING	; bLength, bDescriptorType
			db			'm', 0x00			; bString
			db			'l', 0x00
			db			'k', 0x00
			db			's', 0x00
			db			'r', 0x00
			db			'o', 0x00
			db			'-', 0x00
			db			'h', 0x00
			db			'e', 0x00
			db			'l', 0x00
			db			'i', 0x00
			db			' ', 0x00
			db			' ', 0x00
			db			' ', 0x00
			db			' ', 0x00
			db			' ', 0x00
String2
			db			Descriptor_end-String2, STRING	; bLength, bDescriptorType
			db			'R', 0x00			; bString
			db			'C', 0x00
			db			' ', 0x00
			db			'J', 0x00
			db			'o', 0x00
			db			'y', 0x00
			db			's', 0x00
			db			't', 0x00
			db			'i', 0x00
			db			'c', 0x00
			db			'k', 0x00
			db			' ', 0x00
			db			'V', 0x00
			db			'e', 0x00
			db			'r', 0x00
			db			' ', 0x00
			db			'1', 0x00
Descriptor_end



USBSTUFF	code

InitUSB
			clrf		UIE, ACCESS				; mask all USB interrupts
			clrf		UIR, ACCESS				; clear all USB interrupt flags
			movlw		0x10
			movwf		UCFG, ACCESS			; configure USB for low-speed transfers and to use the on-chip transciever and pull-up resistor
			movlw		0x08
			movwf		UCON, ACCESS			; enable the USB module and its supporting circuitry
			banksel		USB_curr_config
			clrf		USB_curr_config, BANKED
			clrf		USB_idle_rate, BANKED
			clrf		USB_USWSTAT, BANKED		; default to powered state
			movlw		0x01
			movwf		USB_device_status, BANKED
			movwf		USB_protocol, BANKED	; default protocol to report protocol initially
			movlw		NO_REQUEST
			movwf		USB_dev_req, BANKED		; No device requests in process
#ifdef SHOW_ENUM_STATUS
			clrf		TRISB, ACCESS			; set all bits of PORTB as outputs
			movlw		0x01
			movwf		PORTB, ACCESS			; set bit zero to indicate Powered status
#endif
			return

ServiceUSB
			select
				caseset	UIR, UERRIF, ACCESS
					clrf		UEIR, ACCESS
					break
				caseset UIR, SOFIF, ACCESS
					bcf			UIR, SOFIF, ACCESS
					break
				caseset	UIR, IDLEIF, ACCESS
					bcf			UIR, IDLEIF, ACCESS
					bsf			UCON, SUSPND, ACCESS
#ifdef SHOW_ENUM_STATUS
					movlw		0xE0
					andwf		PORTB, F, ACCESS
					bsf			PORTB, 4, ACCESS
#endif
					break
				caseset UIR, ACTVIF, ACCESS
					bcf			UIR, ACTVIF, ACCESS
					bcf			UCON, SUSPND, ACCESS
#ifdef SHOW_ENUM_STATUS
					movlw		0xE0
					andwf		PORTB, F, ACCESS
					banksel		USB_USWSTAT
					movf		USB_USWSTAT, W, BANKED
					select
						case POWERED_STATE
							movlw	0x01
							break
						case DEFAULT_STATE
							movlw	0x02
							break
						case ADDRESS_STATE
							movlw	0x04
							break
						case CONFIG_STATE
							movlw	0x08
					ends
					iorwf		PORTB, F, ACCESS
#endif
					break
				caseset	UIR, STALLIF, ACCESS
					bcf			UIR, STALLIF, ACCESS
					break
				caseset	UIR, URSTIF, ACCESS
					banksel		USB_curr_config
					clrf		USB_curr_config, BANKED
					bcf			UIR, TRNIF, ACCESS		; clear TRNIF four times to clear out the USTAT FIFO
					bcf 		UIR, TRNIF, ACCESS
					bcf			UIR, TRNIF, ACCESS
					bcf			UIR, TRNIF, ACCESS
					banksel		BD0OBC
					movlw		0x08
					movwf		BD0OBC, BANKED
					movlw		low USB_Buffer			; EP0 OUT gets a buffer...
					movwf		BD0OAL, BANKED
					movlw		high USB_Buffer
					movwf		BD0OAH, BANKED			; ...set up its address
					movlw		0x88					; set UOWN bit (USB can write)
					movwf		BD0OST, BANKED
					movlw		low (USB_Buffer+0x08)	; EP0 IN gets a buffer...
					movwf		BD0IAL, BANKED
					movlw		high (USB_Buffer+0x08)
					movwf		BD0IAH, BANKED			; ...set up its address
					movlw		0x08					; clear UOWN bit (MCU can write)
					movwf		BD0IST, BANKED
					clrf		UADDR, ACCESS			; set USB Address to 0
					clrf		UIR, ACCESS				; clear all the USB interrupt flags
					movlw		ENDPT_CONTROL
					movwf		UEP0, ACCESS			; EP0 is a control pipe and requires an ACK
					movlw		0xFF					; enable all error interrupts
					movwf		UEIE, ACCESS
					banksel		USB_USWSTAT
					movlw		DEFAULT_STATE
					movwf		USB_USWSTAT, BANKED
					movlw		0x01
					movwf		USB_device_status, BANKED	; self powered, remote wakeup disabled
#ifdef SHOW_ENUM_STATUS
					movlw		0xE0
					andwf		PORTB, F, ACCESS
					bsf 		PORTB, 1, ACCESS		; set bit 1 of PORTB to indicate Powered state
#endif
					break
				caseset	UIR, TRNIF, ACCESS
					movlw		0x04
					movwf		FSR0H, ACCESS
					movf		USTAT, W, ACCESS
					movwf		FSR0L, ACCESS
					banksel		USB_buffer_desc
					movf		POSTINC0, W
					movwf		USB_buffer_desc, BANKED
					movf		POSTINC0, W
					movwf		USB_buffer_desc+1, BANKED
					movf		POSTINC0, W
					movwf		USB_buffer_desc+2, BANKED
					movf		POSTINC0, W
					movwf		USB_buffer_desc+3, BANKED
					movf		USTAT, W, ACCESS
					movwf		USB_USTAT, BANKED		; save the USB status register
					bcf			UIR, TRNIF, ACCESS		; clear TRNIF interrupt flag
#ifdef SHOW_ENUM_STATUS
					andlw		0x18					; extract EP bits
					select
						case EP0
							movlw		0x20
							break
						case EP1
							movlw		0x40
							break
						case EP2
							movlw		0x80
							break
					ends
					xorwf		PORTB, F, ACCESS		; toggle bit 5, 6, or 7 of PORTB to reflect EP activity
#endif
					movf		USB_buffer_desc, W, BANKED
					andlw		0x3C					; extract PID bits
					select
						case TOKEN_SETUP
							call		ProcessSetupToken
							break
						case TOKEN_IN
							call		ProcessInToken
							break
						case TOKEN_OUT
							call		ProcessOutToken
							break
					ends
					break
			ends
			return

ProcessSetupToken
			banksel		USB_buffer_data
			movf		USB_buffer_desc+ADDRESSH, W, BANKED
			movwf		FSR0H, ACCESS
			movf		USB_buffer_desc+ADDRESSL, W, BANKED
			movwf		FSR0L, ACCESS
			movf		POSTINC0, W
			movwf		USB_buffer_data, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+1, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+2, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+3, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+4, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+5, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+6, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+7, BANKED
			banksel		BD0OBC
			movlw		0x08
			movwf		BD0OBC, BANKED					; reset the byte count
			movwf		BD0IST, BANKED					; return the in buffer to us (dequeue any pending requests)
			banksel		USB_buffer_data+bmRequestType
			ifl	USB_buffer_data+bmRequestType, ==, 0x21
				movlw		0xC8
			otherwise
				movlw		0x88
			endi
			banksel		BD0OST
			movwf		BD0OST, BANKED					; set EP0 OUT UOWN back to USB and DATA0/DATA1 packet according to request type
			bcf			UCON, PKTDIS, ACCESS			; assuming there is nothing to dequeue, clear the packet disable bit
			banksel		USB_dev_req
			movlw		NO_REQUEST
			movwf		USB_dev_req, BANKED				; clear the device request in process
			movf		USB_buffer_data+bmRequestType, W, BANKED
			andlw		0x60							; extract request type bits
			select
				case STANDARD
					call		StandardRequests
					break
				case CLASS
					call		ClassRequests
					break
				case VENDOR
					call		VendorRequests
					break
				default
					bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
			ends
			return

StandardRequests
			movf		USB_buffer_data+bRequest, W, BANKED
			select
				case GET_STATUS
					movf		USB_buffer_data+bmRequestType, W, BANKED
					andlw		0x1F					; extract request recipient bits
					select
						case RECIPIENT_DEVICE
							banksel		BD0IAH
							movf		BD0IAH, W, BANKED
							movwf		FSR0H, ACCESS
							movf		BD0IAL, W, BANKED				; get buffer pointer
							movwf		FSR0L, ACCESS
							banksel		USB_device_status
							movf		USB_device_status, W, BANKED	; copy device status byte to EP0 buffer
							movwf		POSTINC0
							clrf		INDF0
							banksel		BD0IBC
							movlw		0x02
							movwf		BD0IBC, BANKED					; set byte count to 2
							movlw		0xC8
							movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
							break
						case RECIPIENT_INTERFACE
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
									bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
									break
								case CONFIG_STATE
									ifl USB_buffer_data+wIndex, <, NUM_INTERFACES
										banksel		BD0IAH
										movf		BD0IAH, W, BANKED
										movwf		FSR0H, ACCESS
										movf		BD0IAL, W, BANKED				; get buffer pointer
										movwf		FSR0L, ACCESS
										clrf		POSTINC0
										clrf		INDF0
										movlw		0x02
										movwf		BD0IBC, BANKED					; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit																								
									otherwise
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									endi
									break
							ends
							break
						case RECIPIENT_ENDPOINT
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP
									andlw		0x0F								; strip off direction bit
									ifset STATUS, Z, ACCESS							; see if it is EP0
										banksel		BD0IAH
										movf		BD0IAH, W, BANKED				; put EP0 IN buffer pointer...
										movwf		FSR0H, ACCESS
										movf		BD0IAL, W, BANKED
										movwf		FSR0L, ACCESS					; ...into FSR0
										ifset UEP0, EPSTALL, ACCESS
											movlw		0x01
										otherwise
											movlw		0x00
										endi
										movwf		POSTINC0
										clrf		INDF0
										movlw		0x02
										movwf		BD0IBC, BANKED					; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									otherwise
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									endi
									break
								case CONFIG_STATE
									banksel		BD0IAH
									movf		BD0IAH, W, BANKED					; put EP0 IN buffer pointer...
									movwf		FSR0H, ACCESS
									movf		BD0IAL, W, BANKED
									movwf		FSR0L, ACCESS						; ...into FSR0
									movlw		high UEP0							; put UEP0 address...
									movwf		FSR1H, ACCESS
									movlw		low UEP0
									movwf		FSR1L, ACCESS						; ...into FSR1
									banksel		USB_buffer_data+wIndex
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP and...
									andlw		0x0F								; ...strip off direction bit
									ifclr PLUSW1, EPOUTEN, ACCESS
									andifclr PLUSW1, EPINEN, ACCESS
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									otherwise
										ifset PLUSW1, EPSTALL, ACCESS
											movlw		0x01
										otherwise
											movlw		0x00
										endi
										movwf		POSTINC0
										clrf		INDF0
										banksel		BD0IBC
										movlw		0x02
										movwf		BD0IBC, BANKED					; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									endi
									break
								default
									bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							ends
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case CLEAR_FEATURE
				case SET_FEATURE
					movf		USB_buffer_data+bmRequestType, W, BANKED
					andlw		0x1F					; extract request recipient bits
					select
						case RECIPIENT_DEVICE
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case DEVICE_REMOTE_WAKEUP
									ifl USB_buffer_data+bRequest, ==, CLEAR_FEATURE
										bcf			USB_device_status, 1, BANKED
									otherwise
										bsf			USB_device_status, 1, BANKED
									endi
									banksel		BD0IBC
									clrf		BD0IBC, BANKED					; set byte count to 0
									movlw		0xC8
									movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									break
								default
									bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
							ends
							break
						case RECIPIENT_ENDPOINT
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP
									andlw		0x0F								; strip off direction bit
									ifset STATUS, Z, ACCESS							; see if it is EP0
										ifl USB_buffer_data+bRequest, ==, CLEAR_FEATURE
											bcf			UEP0, EPSTALL, ACCESS
										otherwise
											bsf			UEP0, EPSTALL, ACCESS
										endi
										banksel		BD0IBC
										clrf		BD0IBC, BANKED					; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									otherwise
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									endi
									break
								case CONFIG_STATE
									movlw		high UEP0							; put UEP0 address...
									movwf		FSR0H, ACCESS
									movlw		low UEP0
									movwf		FSR0L, ACCESS						; ...into FSR0
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP
									andlw		0x0F								; strip off direction bit
									addwf		FSR0L, F, ACCESS					; add EP number to FSR0
									ifset		STATUS, C, ACCESS
										incf		FSR0H, F, ACCESS
									endi
									ifclr INDF0, EPOUTEN, ACCESS
									andifclr INDF0, EPINEN, ACCESS
										bsf			UEP0, EPSTALL, ACCESS			; set EP0 protocol stall bit to signify Request Error
									otherwise
										ifl USB_buffer_data+bRequest, ==, CLEAR_FEATURE
											bcf			INDF0, EPSTALL, ACCESS
										otherwise
											bsf			INDF0, EPSTALL, ACCESS
										endi
										banksel		BD0IBC
										clrf		BD0IBC, BANKED			; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
									endi
									break
								default
									bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							ends
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case SET_ADDRESS
					ifset USB_buffer_data+wValue, 7, BANKED		; if new device address is illegal, send Request Error
						bsf			UEP0, EPSTALL, ACCESS		; set EP0 protocol stall bit to signify Request Error
					otherwise
						movlw		SET_ADDRESS
						movwf		USB_dev_req, BANKED			; processing a SET_ADDRESS request
						movf		USB_buffer_data+wValue, W, BANKED
						movwf		USB_address_pending, BANKED	; save new address
						banksel		BD0IBC
						clrf		BD0IBC, BANKED				; set byte count to 0
						movlw		0xC8
						movwf		BD0IST, BANKED				; send packet as DATA1, set UOWN bit
					endi
					break
				case GET_DESCRIPTOR
					movwf		USB_dev_req, BANKED				; processing a GET_DESCRIPTOR request
					movf		USB_buffer_data+(wValue+1), W, BANKED
					select
						case DEVICE
							movlw		low (Device-Descriptor_begin)
							movwf		USB_desc_ptr, BANKED
							call		Descriptor				; get descriptor length
							movwf		USB_bytes_left, BANKED
							ifl USB_buffer_data+(wLength+1), ==, 0
							andiff USB_buffer_data+wLength, <, USB_bytes_left
								movf		USB_buffer_data+wLength, W, BANKED
								movwf		USB_bytes_left, BANKED
							endi
							call		SendDescriptorPacket
							break
						case CONFIGURATION
							bcf			USB_error_flags, 0, BANKED
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		low (Configuration1-Descriptor_begin)
									break
								default
									bsf			USB_error_flags, 0, BANKED
							ends
							ifclr USB_error_flags, 0, BANKED
								addlw		0x02				; add offset for wTotalLength
								movwf		USB_desc_ptr, BANKED
								call		Descriptor			; get total descriptor length
								movwf		USB_bytes_left, BANKED
								movlw		0x02
								subwf		USB_desc_ptr, F, BANKED	; subtract offset for wTotalLength
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						case STRING
							bcf			USB_error_flags, 0, BANKED
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		low (String0-Descriptor_begin)
									break
								case 1
									movlw		low (String1-Descriptor_begin)
									break
								case 2
									movlw		low (String2-Descriptor_begin)
									break
								default
									bsf			USB_error_flags, 0, BANKED
							ends
							ifclr USB_error_flags, 0, BANKED
								movwf		USB_desc_ptr, BANKED
								call		Descriptor		; get descriptor length
								movwf		USB_bytes_left, BANKED
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						case HID
							bcf			USB_error_flags, 0, BANKED
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		low (HID1-Descriptor_begin)
									break
								default
									bsf			USB_error_flags, 0, BANKED
							ends
							ifclr USB_error_flags, 0, BANKED
								movwf		USB_desc_ptr, BANKED
								call		Descriptor				; get descriptor length
								movwf		USB_bytes_left, BANKED
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						case REPORT
							bcf			USB_error_flags, 0, BANKED
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		REPORT1_LENGTH
									movwf		USB_bytes_left, BANKED	; set descriptor length
									movlw		low (Report1-Descriptor_begin)
									break
								default
									bsf			USB_error_flags, 0, BANKED
							ends
							ifclr USB_error_flags, 0, BANKED
								movwf		USB_desc_ptr, BANKED
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case GET_CONFIGURATION
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS
					banksel		USB_curr_config
					movf		USB_curr_config, W, BANKED
					movwf		INDF0					; copy current device configuration to EP0 IN buffer
					banksel		BD0IBC
					movlw		0x01
					movwf		BD0IBC, BANKED			; set EP0 IN byte count to 1
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_CONFIGURATION
					ifl USB_buffer_data+wValue, <=, NUM_CONFIGURATIONS
						movf		USB_buffer_data+wValue, W, BANKED
						movwf		USB_curr_config, BANKED
						select
							case 0
								movlw		ADDRESS_STATE
								movwf		USB_USWSTAT, BANKED
#ifdef SHOW_ENUM_STATUS
								movlw		0xE0
								andwf		PORTB, F, ACCESS
								bsf			PORTB, 2, ACCESS
#endif
								break
							default
								movlw		CONFIG_STATE
								movwf		USB_USWSTAT, BANKED
								movlw		0x08
								banksel		BD1IBC
								movwf		BD1IBC, BANKED			; set EP1 IN byte count to 8 
								movlw		low (USB_Buffer+0x10)
								movwf		BD1IAL, BANKED			; set EP1 IN buffer address
								movlw		high (USB_Buffer+0x10)
								movwf		BD1IAH, BANKED
								movlw		0x48
								movwf		BD1IST, BANKED			; clear UOWN bit (PIC can write EP1 IN buffer)
								movlw		ENDPT_IN_ONLY
								movwf		UEP1, ACCESS			; enable EP1 for interrupt in transfers
#ifdef SHOW_ENUM_STATUS
								movlw		0xE0
								andwf		PORTB, F, ACCESS
								bsf			PORTB, 3, ACCESS
#endif
						ends
						banksel		BD0IBC
						clrf		BD0IBC, BANKED			; set byte count to 0
						movlw		0xC8
						movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					otherwise
						bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					endi
					break
				case GET_INTERFACE
					movf		USB_USWSTAT, W, BANKED
					select
						case CONFIG_STATE
							ifl USB_buffer_data+wIndex, <, NUM_INTERFACES
								banksel		BD0IAH
								movf		BD0IAH, W, BANKED
								movwf		FSR0H, ACCESS
								movf		BD0IAL, W, BANKED		; get buffer pointer
								movwf		FSR0L, ACCESS
								clrf		INDF0					; always send back 0 for bAlternateSetting
								movlw		0x01
								movwf		BD0IBC, BANKED			; set byte count to 1
								movlw		0xC8
								movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case SET_INTERFACE
					movf		USB_USWSTAT, W, BANKED
					select
						case CONFIG_STATE
							ifl USB_buffer_data+wIndex, <, NUM_INTERFACES
								movf		USB_buffer_data+wValue, W, BANKED
								select
									case 0									; currently support only bAlternateSetting of 0
										banksel		BD0IBC
										clrf		BD0IBC, BANKED			; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
										break
									default
										bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
								ends
							otherwise
								bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
							endi
							break
						default
							bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					ends
					break
				case SET_DESCRIPTOR
				case SYNCH_FRAME
				default
					bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
					break
			ends
			return

ClassRequests
			movf		USB_buffer_data+bRequest, W, BANKED
			select
				case GET_REPORT
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS			; ...into FSR0
					banksel		BUFFER
					movf		BUFFER, W, BANKED		; copy BUFFER...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		BUFFER+1, W, BANKED		; copy BUFFER+1...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		BUFFER+2, W, BANKED		; copy BUFFER+2...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		BUFFER+3, W, BANKED		; copy BUFFER+3...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		BUFFER+4, W, BANKED		; copy BUFFER+4...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		BUFFER+5, W, BANKED		; copy BUFFER+5...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		BUFFER+6, W, BANKED		; copy BUFFER+6...
					movwf		POSTINC0				; ...to EP0 IN buffer
					movf		BUFFER+7, W, BANKED		; copy BUFFER+7...
					movwf		INDF0					; ...to EP0 IN buffer
					banksel		BD0IBC
					movlw		0x08
					movwf		BD0IBC, BANKED			; set EP0 IN buffer byte count to 8
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_REPORT
					movwf		USB_dev_req, BANKED		; processing a SET_REPORT request
					break
				case GET_PROTOCOL
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS			; ...into FSR0
					banksel		USB_protocol
					movf		USB_protocol, W, BANKED
					movwf		INDF0
					banksel		BD0IBC
					movlw		0x01
					movwf		BD0IBC, BANKED			; set byte count to 1
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_PROTOCOL
					movf		USB_buffer_data+wValue, W, BANKED
					movwf		USB_protocol, BANKED	; update the new protocol value
					banksel		BD0IBC
					clrf		BD0IBC, BANKED			; set byte count to 0
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case GET_IDLE
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS			; ...into FSR0
					banksel		USB_idle_rate
					movf		USB_idle_rate, W, BANKED
					movwf		INDF0
					banksel		BD0IBC
					movlw		0x01
					movwf		BD0IBC, BANKED			; set byte count to 1
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_IDLE
					movf		USB_buffer_data+wValue, W, BANKED
					movwf		USB_idle_rate, BANKED	; update the new idle rate
					banksel		BD0IBC
					clrf		BD0IBC, BANKED			; set byte count to 0
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				default
					bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
			ends
			return

VendorRequests
			movf		USB_buffer_data+bRequest, W, BANKED
			select
				default
					bsf			UEP0, EPSTALL, ACCESS	; set EP0 protocol stall bit to signify Request Error
			ends
			return

ProcessInToken
			banksel		USB_USTAT
			movf		USB_USTAT, W, BANKED
			andlw		0x18		; extract the EP bits
			select
				case EP0
					movf		USB_dev_req, W, BANKED
					select
						case SET_ADDRESS
							movf		USB_address_pending, W, BANKED
							movwf		UADDR, ACCESS
							select
								case 0
									movlw		DEFAULT_STATE
									movwf		USB_USWSTAT, BANKED
#ifdef SHOW_ENUM_STATUS
									movlw		0xE0
									andwf		PORTB, F, ACCESS
									bsf			PORTB, 1, ACCESS
#endif
									break
								default
									movlw		ADDRESS_STATE
									movwf		USB_USWSTAT, BANKED
#ifdef SHOW_ENUM_STATUS
									movlw		0xE0
									andwf		PORTB, F, ACCESS
									bsf			PORTB, 2, ACCESS
#endif
							ends
							break
						case GET_DESCRIPTOR
							call		SendDescriptorPacket
							break
					ends
					break
				case EP1
					break
				case EP2
					break
			ends
			return

ProcessOutToken
			banksel		USB_USTAT
			movf		USB_USTAT, W, BANKED
			andlw		0x18		; extract the EP bits
			select
				case EP0
					movf		USB_dev_req, W, BANKED
					select
						case SET_REPORT
							movlw		NO_REQUEST
							movwf		USB_dev_req, BANKED	; clear device request
							banksel		BD0OAH
							movf		BD0OAH, W, BANKED	; put EP0 OUT buffer pointer...
							movwf		FSR0H, ACCESS
							movf		BD0OAL, W, BANKED
							movwf		FSR0L, ACCESS		; ...into FSR0
							movf		INDF0, W			; get the first byte in the buffer and...
							banksel		LED_states
							movwf		LED_states, BANKED	; ...update the LED states with it
							ifset LED_states, 1, BANKED		; set RA1 to reflect the new state of the CAPS LOCK LED
								bsf			PORTA, 1, ACCESS
							otherwise
								bcf			PORTA, 1, ACCESS
							endi
					ends
					banksel		BD0OBC
					movlw		0x08
					movwf		BD0OBC, BANKED
					movlw		0x88
					movwf		BD0OST, BANKED
					clrf		BD0IBC, BANKED		; set byte count to 0
					movlw		0xC8
					movwf		BD0IST, BANKED		; send packet as DATA1, set UOWN bit
					break
				case EP1
					break
				case EP2
					break
			ends
			return

SendDescriptorPacket
			banksel		USB_bytes_left
			ifl USB_bytes_left, <, 8
				movlw		NO_REQUEST
				movwf		USB_dev_req, BANKED		; sending a short packet, so clear device request
				movf		USB_bytes_left, W, BANKED
			otherwise
				movlw		0x08
			endi
			subwf		USB_bytes_left, F, BANKED
			movwf		USB_packet_length, BANKED
			banksel		BD0IBC
			movwf		BD0IBC, BANKED			; set EP0 IN byte count with packet size
			movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
			movwf		FSR0H, ACCESS
			movf		BD0IAL, W, BANKED
			movwf		FSR0L, ACCESS			; ...into FSR0
			banksel		USB_loop_index
			forlf USB_loop_index, 1, USB_packet_length
				call		Descriptor			; get next byte of descriptor being sent
				movwf		POSTINC0			; copy to EP0 IN buffer, and increment FSR0
				incf		USB_desc_ptr, F, BANKED	; increment the descriptor pointer
			next USB_loop_index
			banksel		BD0IST
			movlw		0x40
			xorwf		BD0IST, W, BANKED		; toggle the DATA01 bit
			andlw		0x40					; clear the PIDs bits
			iorlw		0x88					; set UOWN and DTS bits
			movwf		BD0IST, BANKED
			return

SendBuffer
			banksel		BD1IAH
			movf		BD1IAH, W, BANKED		; put EP1 IN buffer pointer...
			movwf		FSR0H, ACCESS
			movf		BD1IAL, W, BANKED
			movwf		FSR0L, ACCESS			; ...into FSR0
			banksel		BUFFER
			movf		BUFFER, W, BANKED		; copy BUFFER...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		BUFFER+1, W, BANKED		; copy BUFFER+1...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		BUFFER+2, W, BANKED		; copy BUFFER+2...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		BUFFER+3, W, BANKED		; copy BUFFER+3...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		BUFFER+4, W, BANKED		; copy BUFFER+4...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		BUFFER+5, W, BANKED		; copy BUFFER+5...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		BUFFER+6, W, BANKED		; copy BUFFER+6...
			movwf		POSTINC0				; ...to EP1 IN buffer
			movf		BUFFER+7, W, BANKED		; copy BUFFER+7...
			movwf		INDF0					; ...to EP1 IN buffer
			banksel		BD1IBC
			movlw		0x08
			movwf		BD1IBC, BANKED			; set EP1 IN buffer byte count to 8
			movf		BD1IST, W, BANKED		; get EP1 IN status register
			andlw		0x40					; extract the DATA01 bit
			xorlw		0x40					; toggle the DATA01 bit
			iorlw		0x88					; set UOWN and DTS bits
			movwf		BD1IST, BANKED			; send packet
			return

APPLICATION	code
Main
			banksel		COUNTER
			for COUNTER, 0x01, 0x1E				; do nothing for a little while
			next COUNTER

Initialization	

			; interrupts		
			bcf		INTCON,	INT0IF, ACCESS		; clear RB0/INT0 interrupt flag
			bsf		INTCON, INT0IE, ACCESS		; enable RB0/INT0 external interrupt
			bcf		RCON, IPEN, ACCESS			; disable priority levels on interrupts (PIC16CXXX compatibility mode)
			bsf		INTCON2, INTEDG0, ACCESS	; RB0/INT0 interrupt on rising edge
			bsf  	INTCON, GIE, ACCESS			; enable global interrupts

			; ports
			movlw	0x0F
			movwf	ADCON1, ACCESS				; set up all ports pins as digital I/Os

			clrf	PORTA, ACCESS
			clrf	TRISA, ACCESS				; PORTA pins all output

			movlw	B'01000000'
			movwf	PORTB						; PORTB: RB6 high, all others low
			movlw	B'10000001'
			movwf	TRISB						; PORTB: RB0/INT and RB7 as input, all others as output
			bsf		INTCON2, RBPU, ACCESS		; weak pull-up disabled

			; timers
			movlw	B'11000111'
			movwf	T0CON, ACCESS				; Timer0: 8 bit, input Fosc/4 (24MHz/4=6MHz), prescaler 1:256 (10.92 ms)

			movlw	B'00000001'
			movwf	T1CON, ACCESS				; Timer1: two 8 bit operation,  1:1 prescaling, input Fosc/4 (24MHz/4=6Mhz: 10,92ms), enabled

			; others
			clrf	nChPulse, ACCESS			; initialize nChPulse (not automatically initialized)

			movlw	.10
			movwf	SignalPresent, ACCESS		; initialize SignalPresent to 10

			movlw	low TMR_CENTER
			movwf	TMRBUFFER					; initialize TMR_BUFFER low bytes with centered value
			movwf	TMRBUFFER+.1
			movwf	TMRBUFFER+.2
			movwf	TMRBUFFER+.3
			movwf	TMRBUFFER+.4
			movwf	TMRBUFFER+.5
			movwf	TMRBUFFER+.6
			movwf	TMRBUFFER+.7
			movwf	TMRBUFFER+.8
			movwf	TMRBUFFER+.9
			movwf	TMRBUFFER+.10
			movwf	TMRBUFFER+.11

			movlw	high TMR_CENTER
			movwf	TMRBUFFER+.12				; initialize TMR_BUFFER high bytes with centered value
			movwf	TMRBUFFER+.13
			movwf	TMRBUFFER+.14
			movwf	TMRBUFFER+.15
			movwf	TMRBUFFER+.16
			movwf	TMRBUFFER+.17
			movwf	TMRBUFFER+.18
			movwf	TMRBUFFER+.19
			movwf	TMRBUFFER+.20
			movwf	TMRBUFFER+.21
			movwf	TMRBUFFER+.22
			movwf	TMRBUFFER+.23

			movlw	0x00						; initialization value for buffer
			movwf	BUFFER, ACCESS				; initialize BUFFER
			movwf	BUFFER+1, ACCESS
			movwf	BUFFER+2, ACCESS
			movwf	BUFFER+3, ACCESS
			movwf	BUFFER+4, ACCESS
			movwf	BUFFER+5, ACCESS
			movwf	BUFFER+6, ACCESS
			movwf	BUFFER+7, ACCESS

			banksel	LED_states
			clrf	LED_states, BANKED

			; configure usb
			call		InitUSB						; initialize the USB registers and serial interface engine
			repeat
				call		ServiceUSB				; service USB requests...
				banksel		USB_USWSTAT
			until USB_USWSTAT, ==, CONFIG_STATE		; ...until the host configures the peripheral

			; main loop
			repeat

				repeat
					call		ServiceUSB			; service USB requests...
				untilset INTCON, T0IF, ACCESS		; ...until Timer0 goes off
				bcf			INTCON, T0IF, ACCESS	; clear Timer0 interrupt flag

				call		CheckSignal				; check signal for setting led
				call		ProcessChannels			; transfer TMRBUFFER timer data to BUFFER channel data

				banksel		BD1IST
				ifclr BD1IST, UOWN, BANKED			; check to see if the PIC owns the EP1 IN buffer
					call		SendBuffer
				endi

			forever



; ******************************************************************
; ISR
;	Interrupt Service Routine
; ******************************************************************
ISR
	; save context; W, STATUS and BSR are automatically saved on interrupt
	movff	PCLATH,	PCLATH_save		; save PCLATH
	movff	PCLATU,	PCLATU_save		; save PCLATU
	movff	FSR0L, FSR0L_save		; save FSR0L
	movff	FSR0H, FSR0H_save		; save FSR0H

Test_RB0_INT0						; Remember: RB0/INT ext. int. isn't affected by PEIE
	btfsc	INTCON, INT0IE 			; RB0/INT ext. int. (ppm trigger) enabled...
	btfss	INTCON, INT0IF			; ...and RB0/INT ext. int. (ppm trigger) occurred?
	goto	EndISR					; no, nothing to be done here
	bcf		INTCON, INT0IF			; yes, clear ext int flag
	call 	PpmTrigger				; process ppm trigger event

EndISR
	; restore context; W, STATUS and BSR are restored by retfie FAST
	movff	PCLATH_save, PCLATH		; restore PCLATH
	movff	PCLATU_save, PCLATU		; restore PCLATU
	movff	FSR0L_save, FSR0L		; restore FSR0L
	movff	FSR0H_save, FSR0H		; restore FSR0H
	retfie	FAST					; return restoring W, STATUS and BSR



; ******************************************************************
; PpmTrigger
;    Process any PPM Trigger event (RB0/INT external interrupt)
;    Called from the ISR
;
; NOTES on Pulse Position Modulation (PPM) timing (ppm signal from Graupner MC-12 measured with oscilloscope)
;
;   22ms (fixed)  |<-------------------------------------------------------------------------->|
;
;   400us (fixed) |<>| 
;
; ->|<---Init.--->|<-Ch.1->|<-Ch.2->|<-Ch.3->|<-Ch.4->|<-Ch.5->|<-Ch.6->|<-Ch.7->|<---Init.--->|<---
;
;   +--+          +--+     +--+     +--+     +--+     +--+     +--+     +--+     +--+          +--+     
;   |  |          |  |     |  |     |  |     |  |     |  |     |  |     |  |     |  |          |  |
; __|8 |__________|1 |_____|2 |_____|3 |_____|4 |_____|5 |_____|6 |_____|7 |_____|8 |__________|1 |_
;
;                 |<------>| channel pulse (modulated): 1500us for channel = 0%;
;                                                       1100us/1900us for channel = -100%/+100%;
;                                                       900us/2100us for channel = -150%/+150%
;
;   |<----------->| init. pulse (modulated): 22ms minus the duration of 7 channels
;                                            11500us for 7 channels = 0%
;                                            14300us/8700us for 7 channels = -100%/+100%; 
;                                            15700us/7300us for 7 channels = -150%/+150%; 
;
; Obviously each pulse width can be measured between two falling edge or two
; rising edge, so doesn't matter if the radio has a positive ppm signal (as the
; one above) or negative (level 1 and 0 inverted) and doesn't matter if we set
; in OPTION_REG the RB0/INT trigger on the rising edge or the falling edge or 
; if there is an inverting separator stage before RB0/INT input pin or not.
;
; ******************************************************************
PpmTrigger
	bcf			T1CON, TMR1ON, ACCESS		; stop Timer1 just to save data and reset
	movf		TMR1L, W, ACCESS			
	movwf		TMR1L_save, ACCESS			; save TMR1L
	movf		TMR1H, W, ACCESS			
	movwf		TMR1H_save, ACCESS			; save TMR1H
	movf		PIR1, W, ACCESS			
	movwf		PIR1_save, ACCESS			; save PIR1
	clrf		TMR1L, ACCESS				; clear TMR1L
	clrf		TMR1H, ACCESS				; clear TMR1H
	bcf			PIR1, TMR1IF, ACCESS		; clear Timer1 overflow flag
	bsf			T1CON, TMR1ON, ACCESS		; immediately restart Timer1, ready for measuring next pulse width

PpmTrigCheckTiming
	btfsc 		PIR1_save, TMR1IF, ACCESS	; Timer1 overflowed?
	goto 		PpmTrigInitPulse			; yes, Timer1 overflowed: should be an init pulse (a channel pulse should be < 2.1ms)
	movlw		.50							; no, we'll assume: init pulse for Timer1 > 2.1ms, else channel pulse; (50*256/65535)*10.92ms=~2.1ms
	subwf		TMR1H_save, W, ACCESS
	btfsc 		STATUS, C, ACCESS			; Timer1 > 2.1ms?
	goto 		PpmTrigInitPulse			; yes: should be an init pulse

PpmTrigChannelPulse							; no: Timer1 not overflowed and < 2.1ms: should be a channel pulse
	movf		nChPulse, W, ACCESS
	btfsc		STATUS, Z, ACCESS			; nChPulse != 0 ?
	return									; no, nChPulse = 0, simply return, waiting for an init pulse
	xorlw		.13							; yes, check if = 13
	btfss		STATUS, Z, ACCESS			; nChPulse = 13?
	goto		PpmTrigSaveChannel			; no, ok! go to next step
	clrf		nChPulse, ACCESS			; yes, 12 channels already processed, reset nChPulse and return, waiting for an init pulse
	return

; reset nChPulse to 1 and re-initialize channels not received during last ppm pulses series
; this reinitialization of channel not received (and then not stored) is made in case their values are corrupted by 
; spurious ppm impulse (for example while connecting the radio or simply touching the jack) or after connection of a 
; radio with more channels than current
PpmTrigInitPulse							
	movf		nChPulse, W, ACCESS			; nChPulse -> W
	btfsc		STATUS, Z, ACCESS			; nChPulse = 0 ?
	goto		PpmTrigInitPulse2			; yes, all 12 channels received during last ppm pulses series, simply assign 1 to nChPulse and return
PpmTrigInitPulse1							; no, less than 12 channels received during last ppm pulses series, reinitialize the remaining channels with centered values
	lfsr		0, TMRBUFFER				; FSR0 points to first byte of TMRBUFFER
	decf		nChPulse, W, ACCESS			; (nChPulse - 1) -> W: offset for TMRBUFFER
	addwf		FSR0L, F, ACCESS			; FSR0L + (nChPulse - 1) -> FSR0L: add offset to FSR0L
	movlw		low TMR_CENTER
	movwf		INDF0, ACCESS				; initialize low byte with centered value
	movlw		.12
	addwf		FSR0L, F, ACCESS			; FSR0L + 12 -> FSR0L: increment FSR0L by 12 for addressing the high byte
	movlw		high TMR_CENTER
	movwf		INDF0, ACCESS				; initialize high byte with centered value
	incf 		nChPulse, F, ACCESS			; increment nChPulse
	movf		nChPulse, W, ACCESS			; nChPulse -> W
	xorlw		.13							; W xor 13 -> W
	btfss		STATUS, Z, ACCESS			; nChPulse = 13?
	goto		PpmTrigInitPulse1			; no, not finished yet, reinitialize another channel
PpmTrigInitPulse2							; yes, finished; reset nChPulse to 1 and return, ready for receiving first channel pulse
	clrf 		nChPulse, ACCESS
	incf 		nChPulse, F, ACCESS			; Assign 1 to nChPulse
	return

PpmTrigSaveChannel							; we have a correctly numbered (1 <= nChPulse <= 8) channel pulse: save channel data
	lfsr		0, TMRBUFFER				; FSR0 points to first byte of TMRBUFFER
	decf		nChPulse, W, ACCESS			; (nChPulse - 1) -> W: offset for TMRBUFFER
	movff		TMR1L_save, PLUSW0			; store low byte channel data to TMRBUFFER (low bytes go to first 12 bytes of TMRBUFFER)
	addlw		.12							; increment offset in W by 12 (note that FSR0 does not change using PLUSW0)
	movff		TMR1H_save, PLUSW0			; store high byte channel data to TMRBUFFER (high bytes go to last 12 bytes of TMRBUFFER)
	incf 		nChPulse, F, ACCESS			; increment nChPulse, ready for next pulse
	return



; ******************************************************************
; CheckSignal
;
; Note: If using the original Schematic from Alessio's website
; Change the 2 lines from PORTB, 6, ACCESS
; to: PORTB, 5, ACCESS
; As they are currently set to use Malc C's schematic
; ******************************************************************
CheckSignal
	btfsc	PIR1, TMR1IF, ACCESS 		; Timer1 overflowed?
	goto	Tmr1Overflowed				; yes, more than 10.92ms without an RB0/INT trigger event
Tmr1NotOverflowed
	bsf		PORTB, 6, ACCESS			; led on
	movlw	.10
	movwf	SignalPresent, ACCESS		; reset SignalPresent to 10: about 10,92ms * 10 before considering no signal: in fact main loop is usually executed every 10,92ms
	return
Tmr1Overflowed
	movf	SignalPresent, F, ACCESS
	btfsc	STATUS, Z, ACCESS			; SignalPresent is zero?
	goto	NoPpmSignal					; yes, more than 10 consecutive Loop with Tmr1 overflowed: consider no ppm signal
	decf	SignalPresent, F, ACCESS	; no, decrement SignalPresent
	return
NoPpmSignal
	bcf		PORTB, 6, ACCESS			; led off
	return



; ******************************************************************
; ProcessChannels
; 	rescale and store channels to BUFFER
;
; ******************************************************************
ProcessChannels
	movlw	.8
	movwf	nChannel, ACCESS			; initialize nChannel with the number of channels to process
ProcCh_Rescale
	call	ReadChannelAssignment		; read the value of nAssignment corresponding to nChannel
	lfsr	0, TMRBUFFER				; FSR0 points to first byte of TMRBUFFER
	decf	nChannel, W, ACCESS			; (nChannel - 1) -> W: offset for TMRBUFFER
	movff	PLUSW0, TmrL				; save nChannel(th) byte of TMRBUFFER to TmrL
	addlw	.12							; increment offset in W by 12 (note that FSR0 does not change using PLUSW0)
	movff	PLUSW0, TmrH				; save (nChannel+12)(th) byte of TMRBUFFER to TmrH
	call	TmrToChData					; from TmrL and TmrH calculate the rescaled value returning it in ChDataL and ChDataH
ProcCh_Store
ProcCh_Store1
	movf	nAssignment, W, ACCESS
	xorlw	.1
	btfss	STATUS, Z, ACCESS			; nAssignment = 1 ?
	goto	ProcCh_Store2				; no, go to next
	movf	ChDataL, W, ACCESS			; yes, we have to store this channel data; ChDataL -> W
	movwf	BUFFER, ACCESS				; save ChDataL to BUFFER
	goto	ProcCh_CheckFinish			; check if all channels processed
ProcCh_Store2
	movf	nAssignment, W, ACCESS
	xorlw	.2
	btfss	STATUS, Z, ACCESS			; nAssignment = 2 ?
	goto	ProcCh_Store3				; no, go to next
	movf	ChDataL, W, ACCESS			; yes, we have to store this channel data; ChDataL -> W
	movwf	BUFFER+1, ACCESS			; save ChDataL to BUFFER
	goto	ProcCh_CheckFinish			; check if all channels processed
ProcCh_Store3
	movf	nAssignment, W, ACCESS
	xorlw	.3
	btfss	STATUS, Z, ACCESS			; nAssignment = 3 ?
	goto	ProcCh_Store4				; no, go to next
	movf	ChDataL, W, ACCESS			; yes, we have to store this channel data; ChDataL -> W
	movwf	BUFFER+2, ACCESS				; save ChDataL to BUFFER
	goto	ProcCh_CheckFinish			; check if all channels processed
ProcCh_Store4
	movf	nAssignment, W, ACCESS
	xorlw	.4
	btfss	STATUS, Z, ACCESS			; nAssignment = 4 ?
	goto	ProcCh_Store5				; no, go to next
	movf	ChDataL, W, ACCESS			; yes, we have to store this channel data; ChDataL -> W
	movwf	BUFFER+3, ACCESS				; save ChDataL to BUFFER
	goto	ProcCh_CheckFinish			; check if all channels processed
ProcCh_Store5
	movf	nAssignment, W, ACCESS
	xorlw	.5
	btfss	STATUS, Z, ACCESS			; nAssignment = 5 ?
	goto	ProcCh_Store6				; no, go to next
	movf	ChDataL, W, ACCESS			; yes, we have to store this channel data; ChDataL -> W
	movwf	BUFFER+4, ACCESS				; save ChDataL to BUFFER
	goto	ProcCh_CheckFinish			; check if all channels processed
ProcCh_Store6
	movf	nAssignment, W, ACCESS
	xorlw	.6
	btfss	STATUS, Z, ACCESS			; nAssignment = 6 ?
	goto	ProcCh_Store7				; no, go to next
	movf	ChDataL, W, ACCESS			; yes, we have to store this channel data; ChDataL -> W
	movwf	BUFFER+5, ACCESS			; save ChDataL to BUFFER
	goto	ProcCh_CheckFinish			; check if all channels processed
ProcCh_Store7
	movf	nAssignment, W, ACCESS
	xorlw	.7
	btfss	STATUS, Z, ACCESS			; nAssignment = 7 ?
	goto	ProcCh_Store8				; no, go to next
	movf	ChDataL, W, ACCESS			; yes, we have to store this channel data; ChDataL -> W
	movwf	BUFFER+6, ACCESS				; save ChDataL to BUFFER
	goto	ProcCh_CheckFinish			; check if all channels processed
ProcCh_Store8
	movf	nAssignment, W, ACCESS
	xorlw	.8
	btfss	STATUS, Z, ACCESS			; nAssignment = 8 ?
	goto	ProcCh_CheckFinish				; no, go to next
	movf	ChDataL, W, ACCESS			; yes, we have to store this channel data; ChDataL -> W
	movwf	BUFFER+7, ACCESS				; save ChDataL to BUFFER
	goto	ProcCh_CheckFinish			; check if all channels processed
ProcCh_CheckFinish
	decfsz	nChannel,F, ACCESS			; decrement nChannel, result is 0 ?
	goto	ProcCh_Rescale				; no, continue with next channel; 
	return								; yes, return: all 8 channels processed



; ******************************************************************
; TmrToChData
; 	The following procedure use an algorithm for scaling the ~5000 points of the channel timer range
; 	to the reduced output 2^NBITS points trying to use the full NBITS output resolution; the
; 	algorithm is made of a number of step equal to the number of output bits NBITS. The input is
; 	in TmrL, TmrH; the output in ChDataL, ChDataH.
; ******************************************************************
;TmrToChData
;	; in this case (4 analog channels and 8 buttons) we use 14 bits of output resolution for each
;	; analog channel: the scaling mentioned above is not necessary; we directly transfer Tmr to
;	; ChData without any manipulation and leave the rest to the HID driver by properly setting
;	; logical and physical minimum and maximum in the HID report descriptor
;	movff	TmrL, ChDataL
;	movlw	0x00
;	movwf	TmrH,ACCESS
;	;movff	TmrH, ChDataH
TmrToChData
	; initialize StepNumber = NBITS - 1: number of cycles in the loop below
	movlw		NBITS - 1
	movwf		StepNumber, ACCESS
	; NBITS(th) cycle made outside loop initializing Step and Ref:
	; initialize Step with a value as close as possible to the real Tmr range (Tmr(+100%)-Tmr(-100%)) but
	; multiple of 2^NBITS (and than exactly divisible by 2 at least NBITS times), dividing it by 2 for the first time
	; see usb_defs.inc for the definition and meaning of TMR_RANGE
	movlw		low (TMR_RANGE / 2)
	movwf		StepL, ACCESS			; Initialize StepL
	movlw		high (TMR_RANGE / 2)
	movwf		StepH, ACCESS			; Initialize StepH
	; initialize Ref with the Tmr value corresponding to a channel centered (Tmr(0%))
	; see usb_defs.inc for the definition and meaning of TMR_CENTER
	movlw		low TMR_CENTER
	movwf		RefL, ACCESS			; Initialize RefL
	movlw		high TMR_CENTER
	movwf		RefH, ACCESS			; Initialize RefH
	; initilize ChData to 0
	clrf		ChDataL, ACCESS			; clear ChDataL
	clrf		ChDataH, ACCESS			; clear ChDataH
	; calculate ChannelData MSb
	call		TmrNotLessThanRef		; Tmr >= Ref = Tmr(0%) ? result on STATUS,C
	rlcf		ChDataL, F, ACCESS		; STATUS,C -> ChData LSb, it will be the MSb (NBITS(th) bit) at the end of the loop
TmrToChDataLoop; NBITS - 1 steps loop, each one calculate one of the remaining bit of ChData to fill
	; divide step by 2
	bcf			STATUS, C, ACCESS		; clear carry bit to fill StepH MSb with a 0 on rotating
	rrcf		StepH, F, ACCESS		; divide Step by 2; StepH LSb -> STATUS,C -> ...
	rrcf		StepL, F, ACCESS		; ... -> StepL MSb
	; translate Ref adding/subtracting Step according to last Tmr - Ref comparing result: >= add, < subtract
	btfss		ChDataL, 0, ACCESS		; Tmr was >= Ref ? 
	goto		TmrToChDataLoop2		; no, subtract Step from Ref
TmrToChDataLoop1						; yes, add Step to Ref
	movf		StepL, W, ACCESS
	addwf		RefL, F, ACCESS			; RefL + StepL -> RefL
	btfsc		STATUS, C, ACCESS		; RefL overflowed ?
	incf		RefH, F, ACCESS			; yes, increment RefH; suppose RefL RefH to be in the same bank
	movf		StepH, W, ACCESS
	addwf		RefH, F, ACCESS			; RefH + StepH -> RefL; can't overflow if Tmr/Step initializations are correct
	goto		TmrToChDataLoop3		; proceed with next phase
TmrToChDataLoop2						; subtract Step from Ref
	movf		StepL, W, ACCESS
	subwf		RefL, F, ACCESS			; RefL - StepL -> RefL
	btfss		STATUS, C, ACCESS		; RefL was < StepL ?
	decf		RefH, F, ACCESS			; yes, decrement RefH; suppose RefL RefH to be in the same bank
	movf		StepH, W, ACCESS
	subwf		RefH, F, ACCESS			; RefH - StepH -> RefL; can't be RefH < StepH if Tmr/Step initializations are correct
TmrToChDataLoop3						; final phase of the loop: calculate ChData using the new Ref
	; calculate ChannelData 
	call		TmrNotLessThanRef		; Tmr >= Ref ? result on STATUS,C
	rlcf		ChDataL, F, ACCESS		; STATUS,C -> ChDataL LSb; ChDataL MSb -> STATUS,C -> ...
	rlcf		ChDataH, F, ACCESS		; ... -> ChDataH LSb
	; check if loop is finished
	decfsz		StepNumber, F, ACCESS	; decrement StepNumber, result is 0 ?
	goto		TmrToChDataLoop			; no, continue loop
	goto		ReadChannelAssignment
	return								; yes, all steps completed: done

; ******************************************************************
; TmrNotLessThanRef
;	Tmr >= Ref ? result on STATUS,C
; ******************************************************************
TmrNotLessThanRef
	movf		RefH, W, ACCESS
	subwf		TmrH, W, ACCESS			; TmrH - RefH -> W
	btfss		STATUS, C, ACCESS		; TmrH >= RefH ?
	return								; no, return with STATUS,C = 0 (Tmr < Ref)
	btfss		STATUS, Z, ACCESS		; yes, TmrH = RefH ?
	return								; no, return with STATUS,C = 1 (Tmr > Ref)
	movf		RefL, W, ACCESS			; yes, see low bytes too
	subwf		TmrL, W, ACCESS			; TmrL - RefL -> W
	return								; no, return with STATUS,C = 1 if >=, STATUS,C = 0 if <




; ******************************************************************
; ReadChannelAssignment
;	read nAssignment for the current nChannel from the
;	proper assignment table, according to jumper settings
; ******************************************************************
ReadChannelAssignment
;	btfsc		PORTB, 7, ACCESS		; RB7 low
;	goto		ReadChAssignFutaba		; no, Futaba radio
ReadChAssignJR							; yes, JR radio
	; prepare table pointer
;	movlw		low JRAssignment
;	movwf		TBLPTRL, ACCESS
;	movlw		high JRAssignment
;	movwf		TBLPTRH, ACCESS
;	movlw		upper JRAssignment
;	movwf		TBLPTRU, ACCESS			;TBLPTR points to first byte of table
;	goto		ReadAssignTable
ReadChAssignFutaba
	; prepare table pointer
	movlw		low FutabaAssignment
	movwf		TBLPTRL, ACCESS
	movlw		high FutabaAssignment
	movwf		TBLPTRH, ACCESS
	movlw		upper FutabaAssignment
	movwf		TBLPTRU, ACCESS			;TBLPTR points to first byte of table
	goto		ReadAssignTable
ReadAssignTable
	; add offset (nChannel - 1) to pointer
	decf	nChannel, W, ACCESS			; nChannel - 1 -> W
	addwf	TBLPTRL, F, ACCESS			; TBLPTRL + iKeyL -> TBLPTRL
	movlw	0							; 0 -> W
	addwfc	TBLPTRH, F, ACCESS			; TBLPTRH + 0 + C -> TBLPTRH
	addwfc	TBLPTRU, F, ACCESS			; TBLPTRU + 0 + C -> TBLPTRU
	tblrd*								; table -> TABLAT
	movf	TABLAT, W, ACCESS			; TABLAT -> W
	movwf	nAssignment, ACCESS
	return


; assignment tables: ordered by channel number: 1..8
; assignments: joystick axes: 1..8

; assignment table for JR/Graupner
JRAssignment
	db		.3, .1		; ch 1 -> axis 3 (throttle),  ch 2 -> axis 1 (aileron)
	db		.2, .4		; ch 3 -> axis 2 (elevator),  ch 4 -> axis 4 (rudder)
	db		.5, .6		; ch 5 -> axis 5,  ch 6 -> axis 6
	db		.7, .8		; ch 7 -> axis 7,  ch 8 -> axis 8
	db		.9, .10		; ch 9 -> axis 9,  ch 10 -> axis 10
	db		.11, .12	; ch 11 -> axis 11,  ch 12 -> axis 12

; assignment table for Futaba/Hitec
FutabaAssignment
	db		.1, .2		; ch 1 -> axis 1 (aileron),  ch 2 -> axis 2 (elevator)
	db		.3, .4		; ch 3 -> axis 3 (throttle),  ch 4 -> axis 4 (rudder)
	db		.5, .6		; ch 5 -> axis 5,  ch 6 -> axis 6
	db		.7, .8		; ch 7 -> axis 7,  ch 8 -> axis 8
	db		.9, .10		; ch 9 -> axis 9,  ch 10 -> axis 10
	db		.11, .12	; ch 11 -> axis 11,  ch 12 -> axis 12

	end
